from pprint import pprint
import boto3
from boto3.dynamodb.conditions import Key
from boto3.dynamodb.conditions import Attr
from botocore.exceptions import ClientError
from decimal import Decimal

def put_item(item_dict, table_name, dynamodb):
  '''
  Inserts json item into DynamoDB table
  item_dict = {
    "attr" : "value",
    "attr2" : "value2"
  }
  table_name = "alerts
  '''
  table = dynamodb.Table(table_name)
  response = 'None'
  try:
    response = table.put_item(
        Item = item_dict
    )
  except ClientError as e:
    print(e.response['Error']['Message'])
  

  return response

def get_item(key_dict, table_name, dynamodb):
  '''
  Retrieves item from DynamoDB table
  key_dict = {
    "prim_key" = "value",
    "sort_key" = "value"
  }
  '''
  table = dynamodb.Table(table_name)
  try:
    response = table.get_item(Key=key_dict)
  except ClientError as e:
    print(e.response['Error']['Message'])
  else:
    return response['Item']

def update_item(key_dict, update_expression, expression_attr_values, table_name, dynamodb):
  '''
  Retrieves item from DynamoDB table
  key_dict = {
    "prim_key" = "value",
    "sort_key" = "value"
  }
  update_expression = "set service_graph=:i, metric_list=:l, significance_score=:s"
  expression_attr_values = {
    ':i': {'s1':['s2', 's3']},
    ':l': ['124','123'],
    ':s': Decimal(35.5)
  }
  #example to append to list
  UpdateExpression="SET some_attr = list_append(if_not_exists(some_attr, :empty_list), :i)",
  ExpressionAttributeValues={
    ':i': [some_value],
    "empty_list" : []
  }
  '''
  table = dynamodb.Table(table_name)
  response = table.update_item(
    Key = key_dict,
    UpdateExpression= update_expression,
    ExpressionAttributeValues= expression_attr_values,
    ReturnValues="UPDATED_NEW"
  )
  return response

def update_item_conditionally(key_dict, condition_expression, update_expression, expression_attr_values, table_name, dynamodb):
  '''
  Retrieves item from DynamoDB table
  key_dict = {
    "prim_key" = "value",
    "sort_key" = "value"
  }
  update_expression = "set service_graph=:i, metric_list=:l, significance_score=:s"
  expression_attr_values = {
    ':i': {'s1':['s2', 's3']},
    ':l': ['124','123'],
    ':s': Decimal(35.5)
  }
  condition_expression = "significance_score <= :val"
  '''
  table = dynamodb.Table(table_name)
  response = table.update_item(
    Key = key_dict,
    UpdateExpression= update_expression,
    ExpressionAttributeValues= expression_attr_values,
    ConditionExpression = condition_expression,
    ReturnValues="UPDATED_NEW"
  )
  return response

def delete_item_conditionally(key_dict, condition_expression, expression_attr_values, table_name, dynamodb):
  '''
  condition_expression = "significance_score <= :val"
  expression_attr_values = {
    ":val": Decimal(50)
  }
  key_dict = {
    'org_id': 'Aptoide',
    'start_time': '2020-09-03 12:00:00'
  }
  '''
  table = dynamodb.Table(table_name)
  try:
    response = table.delete_item(
      Key=key_dict,
      ConditionExpression = condition_expression,
      ExpressionAttributeValues = expression_attr_values,
    )
  except ClientError as e:
    if e.response['Error']['Code'] == "ConditionalCheckFailedException":
      print(e.response['Error']['Message'])
    else:
      raise
  else:
    return response

def query_by_key(key_condition, table_name, dynamodb):
  '''
  Queries from DynamoDB table by key condition
  key_condition = Key('org_id').eq('Aptoide')
  '''
  table = dynamodb.Table(table_name)
  try:
    response = table.query(
      KeyConditionExpression=key_condition
    )
  except ClientError as e:
    print(e.response['Error']['Message'])
  else:
    return response['Items']

def query_and_project_by_key_condition(projection_expr, expr_attr_names, key_condition, table_name, dynamodb):
  '''
  Queries from DynamoDB table by key condition and only returns some attrs
  key_condition = Key('year').eq(year) & Key('title').between(title_range[0], title_range[1])
  projection_expr = "#yr, title, info.genres, info.actors[0]"
  expr_attr_names = {"#yr": "year"}
  '''
  table = dynamodb.Table(table_name)
  try:
    response = table.query(
      KeyConditionExpression = key_condition,
      ProjectionExpression = projection_expr,
      ExpressionAttributeNames = expr_attr_names
    )
  except ClientError as e:
    print(e.response['Error']['Message'])
  else:
    return response['Items']

def scan_table(scan_kwargs, table_name, dynamodb):
  '''
  Scans entire table looking for items that match the filter expression
  scan_kwargs = {
    'FilterExpression': Key('year').between(*year_range),
    'ProjectionExpression': "#yr, title, info.rating",
    'ExpressionAttributeNames': {"#yr": "year"}
  }
  '''
  table = dynamodb.Table(table_name)
  done = False
  start_key = None
  result_list = []
  while not done:
    if start_key:
      scan_kwargs['ExclusiveStartKey'] = start_key
    response = table.scan(**scan_kwargs)
    #Do SMTH with response
    result_list.append(response['Items'])
    start_key = response.get('LastEvaluatedKey', None)
    done = start_key is None
  return result_list

def query_by_key_min_max(key_condition, table_name, is_min, dynamodb):
  '''
  Queries from DynamoDB table by key condition
  key_condition = Key('part_id').eq(partId) & Key('range_key').between(start, end)
  '''
  table = dynamodb.Table(table_name)
  try:
    response = table.query(
      KeyConditionExpression=key_condition,
      ScanIndexForward = is_min,
      Limit = 1
    )
  except ClientError as e:
    print(e.response['Error']['Message'])
  else:
    return response['Items']



from boto3.dynamodb.conditions import Key
from boto3.dynamodb.conditions import Attr

from .utils.dynamodb_utils import put_item, delete_item_conditionally
from .utils.dynamodb_utils import get_item, query_by_key, \
  query_and_project_by_key_condition, scan_table, query_by_key_min_max
from .utils.dynamodb_utils import update_item, update_item_conditionally


ALERTS_TABLE = 'alerts.items'
ALERTS_CONFIGS_TABLE = 'alerts.config'
METRICS_TIMESERIES_TABLE = 'metric.ts'
LOGS_METRICS_FETCHING_TABLE = 'logs.cloud_metric_fetching'

############ ALERTS ###########################
def insert_alert(alert_id, metric_id, org_id, app_id, team_id, assigned_to, start_time,
  end_time, alert_description, is_acknowledged, anomalies_dict, related_prev_anomalies,
  service_graph, significance_score, dynamodb):
  '''
  insert_alert(alert_id = "256828", metric_id = 123, org_id = 'EDP', app_id = 'EDP Online', team_id = 'EDP Online', assigned_to = 'Jorge', \
  start_time = '2020-09-03 12:00:00', end_time = '2020-09-03 12:20:00', alert_description = 'Spike in costs',\
  is_acknowledged = 'True', anomalies_dict = {}, related_prev_anomalies = {},
  service_graph = {}, significance_score = '34.3')
  '''
  alert_dict = {
    'alert_id' : alert_id,
    'metric_id' : metric_id,
    'org_id': org_id,
    'app_id': app_id,
    'team_id' : team_id,
    'assigned_to' : assigned_to,
    'start_time' : start_time,
    'end_time' : end_time,
    'alert_description' : alert_description,
    'is_acknowledged' : is_acknowledged,
    'anomalies_dict' : anomalies_dict,
    'related_prev_anomalies' : related_prev_anomalies,
    'service_graph': service_graph,
    'significance_score':  significance_score
  }

  print(alert_dict)
  response = put_item(alert_dict, ALERTS_TABLE, dynamodb)
  return response

def get_alert_item_by_key(anom_id, dynamodb):
  key_dict = {
    "alert_id" : anom_id, 
  }
  alert_item = get_item(key_dict, ALERTS_TABLE, dynamodb)
  return alert_item

def update_alert_with_related_anomalies(alert_id,start_time, corr_anoms_dict, related_prev_anomalies, dynamodb):
  #https://stackoverflow.com/questions/52367094/how-to-update-dynamodb-table-with-dict-data-type-boto3
  #TODO add similar alerts
  update_expression = "set anomalies_dict=:a, related_prev_anomalies=:r"
  expression_attr_values = {
    ':a': corr_anoms_dict,
    ':r': related_prev_anomalies,
  }
  key_dict = {
    'alert_id' : alert_id,
    'start_time' : start_time,
  }
  
  update_item(key_dict, update_expression, expression_attr_values, ALERTS_TABLE, dynamodb)

  return True 

def terminate_alert(alert_id,start_time, end_timestamp, dynamodb):
  update_expression = "set end_timestamp=:e"
  expression_attr_values = {
    ':e': end_timestamp
  }
  key_dict = {
    'alert_id' : alert_id,
    'start_time' : start_time,
  }
  
  update_item(key_dict, update_expression, expression_attr_values, ALERTS_TABLE, dynamodb)

  return True 


############ METRICS ##########################
#create new metric_timeseries item for current day
def create_metric(metric_id, date_bucket, anom_alarm_id, metric_name, provider, namespace,
  agent, org_id, app_id, alignment, groupby, first, last, data_points_list, dynamodb):
  '''
  create_metric(
    metric_id = 123, date_bucket = "2020-09-10",
    anom_alarm_id = "crazy_monkey.juicy_bananas.kiwi_count", metric_name = "error_rate",
    provider = "aws", namespace = "dynamodb", agent = "CloudWatch", org_id = "Aptoide",
    app_id = "app1", alignment = "DELTA", groupby = "service", first = 1535530412,
    last = 1535530432, data_points_list = [
          { 'val': 50, 'time': 1535530412},
	        { 'val': 55, 'time' : 1535530415},
	        { 'val': 56, 'time': 1535530420},
	        { 'val': 55, 'time' : 1535530430}, 
	        { 'val': 56, 'time': 1535530432}
      ]
  )
  '''
  metric_dict = {
    "metric_id" : metric_id, 
    "date_bucket" : date_bucket,
    "metric_name" : metric_name,
    "encrypted_id" : anom_alarm_id,
    "provider" : provider,
    "namespace" : namespace,
    "agent" : agent,
    "org_id" : org_id,
    "app_id" : app_id, 
    "alignment" : alignment,
    "groupby" : groupby,
    "first" : first,
    "last": last,
    "data_points_list" : data_points_list
  }

  #print(alert_dict)
  response = put_item(metric_dict, METRICS_TIMESERIES_TABLE, dynamodb)
  return response

#query metrics info by metric_id key
def get_metric_item_by_key(metric_id, curr_date, dynamodb):
  #get_metric_item_by_key(123, '2020-09-10')
  key_dict = {
    "metric_id" : metric_id, 
    "date_bucket" : str(curr_date)
  }

  metric_item = get_item(key_dict, 'metric.ts', dynamodb)
  return metric_item

def scan_metrics_by_encrypted_id(anom_alarm_id, dynamodb):
  scan_kwargs = {
    'FilterExpression': Attr('encrypted_id').eq(anom_alarm_id),
  }
  metric_list = scan_table(scan_kwargs, 'metric.ts', dynamodb)
  metric_item = metric_list[0]
  return metric_item


############## ALERT CONFIGS ################
def query_alerts_configs_by_key(metric_id, dynamodb):
  key_condition = Key('metric_id').eq(metric_id)
  alerts_configs_list = query_by_key(key_condition, ALERTS_CONFIGS_TABLE, dynamodb)
  return alerts_configs_list  

def insert_alert_config(metric_id, alert_title, severity, alert_type, alert_direction, description, 
  duration, duration_unit, rule_dict, recipients_list, owner_dict, dynamodb):
  '''
  insert_alert_config(
    metric_id = "metric1245", alert_title = "Anomaly by Cluster", severity = "critical",
    alert_type = "anomaly", alert_direction = "spikes/drops", description = "Relevant to Play Store billing user journey",
    duration= 12, duration_unit = "hours", rule_dict = {}, recipients_list = [{
      "channel" : "webhook", 
      "contact" : "j.velez2210@gmail.com"
      },{
        "channel" : "slack",
        "contact" : "j.velez2210@gmail.com"
      }
    ], 
    owner_dict = {
      "user_id" : "user12341",
      "user_name" : "João Tótó",
    }
  )
  '''
  alert_dict = {
    "metric_id" : metric_id, 
    "alert_title" : alert_title,
    "severity" : severity, 
    "alert_type" : alert_type,
    "alert_direction" : alert_direction, 
    "description" : description,
    "duration" : duration,
    "duration_unit" : duration_unit,
    "rule_dict" : rule_dict,
    "recipients" : recipients_list,
    "owner" : owner_dict
  }

  #print(alert_dict)
  response = put_item(alert_dict, 'alerts.config')
  return response

############## LOGS #########################
def query_most_recent_metric_fetching_log(component_id, dynamodb):
  '''
  Fetches the log with the highest timestamp, from all the logs between start & end ts
  '''
  #TODO Insert a greater than last_fetched_ts condition to prevent the query to go too far in time
  key_condition = Key('component_id').eq(component_id) #& Key('last_fetched_ts').between(start_ts, end_ts)
  logs_list = query_by_key_min_max(key_condition, LOGS_METRICS_FETCHING_TABLE, False, dynamodb)
  log_item = logs_list[0]
  print(log_item)
  return log_item

############## EXAMPLES #####################
def update_values_in_dict_attr():
  '''
  dynamo = boto3.resource('dynamodb')  
    tbl = dynamo.Table('<TableName>')  

    result = tbl.update_item(  
        Key={  
            "game_id": game_id  
        },  
        UpdateExpression="SET players.#player_id.score = :score_val",  
        ExpressionAttributeNames={  
            "#player_id": player_id  
        },  
        ExpressionAttributeValues={  
            ":score_val": score_val  
        }  
    )
  '''
  return True


########## Function Calls ##########################
#query_most_recent_metric_fetching_log('prd_db')